/******************************************************************************
 *  @author Robert Sedgewick
 *  @author Kevin Wayne
 *
 *  Compilation:  javac BST.java
 *  Execution:    java BST
 *
 *  A symbol table implemented with a binary search tree.
 *
 ******************************************************************************/

import java.util.NoSuchElementException;

/**
 * Represents an ordered symbol table of generic key-value pairs. 
 * This implementation uses an (unbalanced) binary search tree.
 */
public class BST<Key extends Comparable<Key>, Value> {
    private Node root;             // root of BST

    private class Node {
        private Key key;           // sorted by key
        private Value val;         // associated data
        private Node left, right;  // left and right subtrees
        private int count;         // number of nodes in subtree

        public Node(Key key, Value val, int count) {
            this.key = key;
            this.val = val;
            this.count = count;
        }
    }


    // Returns true if this symbol table is empty.
    public boolean isEmpty() { return size() == 0; }

    // Returns the number of key-value pairs in this symbol table.
    public int size() { return size(root); }

    // return number of key-value pairs in BST rooted at x
    private int size(Node x) {
        if (x == null) return 0;
        else return x.count;
    }

    // Does this symbol table contain the given key?
    public boolean contains(Key key) {
        if (key == null) throw new IllegalArgumentException("argument to contains() is null");
        return get(key) != null;
    }

    // Returns the value associated with the given key.
    public Value get(Key key) { return get(root, key); }

    private Value get(Node x, Key key) {
        if (key == null) throw new IllegalArgumentException("calls get() with a null key");
        if (x == null) return null;
        int cmp = key.compareTo(x.key);
        if      (cmp < 0) return get(x.left, key);
        else if (cmp > 0) return get(x.right, key);
        else              return x.val;
    }

    /**
     * Inserts the specified key-value pair into the symbol table, 
     * overwriting the old value with the new value if the symbol table already 
     * contains the specified key.
     */
    public void put(Key key, Value val) {
        if (key == null || val == null) 
            throw new IllegalArgumentException("calls put() with a null key");
        root = put(root, key, val);
    }

    private Node put(Node x, Key key, Value val) {
        if (x == null) return new Node(key, val, 1);
        int cmp = key.compareTo(x.key);
        if      (cmp < 0) x.left  = put(x.left,  key, val);
        else if (cmp > 0) x.right = put(x.right, key, val);
        else              x.val   = val;
        x.count = 1 + size(x.left) + size(x.right);
        return x;
    }  

    /**
     * Return the number of keys in the symbol table strictly less than {@code key}.
     * @throws IllegalArgumentException if {@code key} is {@code null}
     */
    public int rank(Key key) {
        
        // ----- WRITE YOUR CODE HERE ----- //
        // HINT: DO YOU NEED A HELPER METHOD?

        return 0;  
    }
     

    /**
     * Returns the number of keys in the symbol table in the given range.
     * @throws IllegalArgumentException if either {@code lo} or {@code hi}
     *         is {@code null}
     */
    public int rangeCount(Key lo, Key hi) {

        // ----- WRITE YOUR CODE HERE ----- //
        // HINT: CAN YOU USE rank()?

        return 0;  
    }

    public static void main(String[] args) { 
        int[] tree = new int[] {4, 
                                2, 7, 
                                1, 3, 5, 10};
        BST<Integer, Integer> st = new BST<Integer, Integer>();
        for (int i = 0; i < tree.length; i++) {
            st.put(tree[i], i);
        }

        System.out.println("Element\tRank");
        for (int i = 0; i < 9; i++)
            System.out.println(i + "\t" + st.rank(i));

        for (int i = 0; i <= tree.length / 2; i++)
            System.out.println("There are " + 
                                st.rangeCount(tree[i], tree[tree.length-1-i]) +
                                " elements between " + tree[i] + " and " + 
                                tree[tree.length-1-i]);
    }
}
